﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.ServiceModel;
using System.Windows.Forms;
using PeerWcfComm;

namespace UserPeerApplication
{
    /// <summary>
    /// Definition of both the User peer and the main form
    /// for this application. This peer participates in IWorker
    /// communication.
    /// </summary>
    [ServiceBehavior(InstanceContextMode = InstanceContextMode.Single)]
    public partial class User : Form, IWorker
    {
        private ServiceHost _host;
        private ChannelFactory<IWorker> _factory;
        private IWorker _clientChannel;
        private Dictionary<Guid, WorkerStatus> _workers;
        private Random _rnd;

        public User()
        {
            InitializeComponent();

            _rnd = new Random(DateTime.Now.Millisecond);
            _workers = new Dictionary<Guid, WorkerStatus>();

            _host = new ServiceHost(this);
            _host.Open();

            _factory = new ChannelFactory<IWorker>("IWorkerEndpoint");
            _clientChannel = _factory.CreateChannel();

            PeerNode pn = ((IClientChannel)_clientChannel).GetProperty<PeerNode>();
            pn.MessagePropagationFilter = new RemoteOnlyMessagePropagationFilter();
        }

        #region IWorker Members

        void IWorker.Heartbeat(WorkerStatus status)
        {
            lock (_workers)
                _workers[status.WorkerId] = status;
        }

        void IWorker.DoWork(Guid workerId, WorkItem work)
        {
            //This doesn't make sense for a user peer.
        }
        #endregion

        /// <summary>
        /// Tick event handler for our form timer. This will update
        /// the UI according to our latest state.
        /// </summary>
        private void timer1_Tick(object sender, EventArgs e)
        {
            this.lstWorkers.Items.Clear();

            //Copy the Ids of our workers to a temp variable.
            List<Guid> workers;
            lock (_workers)
                workers = new List<Guid>(_workers.Keys);

            //Loop through the list, get the WorkerStatus,
            //and create a ListViewItem for our UI.
            foreach (Guid worker in workers)
            {
                WorkerStatus status;

                lock (_workers)
                    status = _workers[worker];

                this.lstWorkers.Items.Add(
                    new ListViewItem(
                    new string[] { 
                        worker.ToString(), 
                        status.State.ToString(), 
                        status.CurrentWorkItem.ToString(), 
                        status.PendingWorkItemCount.ToString() }));
            }
        }

        private void btnNewWork_Click(object sender, EventArgs e)
        {
            //Create a new WorkItem with random Id and a
            //"work" period from 0 to 15 seconds.
            WorkItem wrkItem = new WorkItem();
            wrkItem.WorkItemId = Guid.NewGuid();
            wrkItem.PretendToWorkSeconds = _rnd.Next(15);

            //Get the next worker that "needs" work.
            WorkerStatus nextWorker = GetNextWorker();

            if (nextWorker != null)
            {
                _clientChannel.DoWork(nextWorker.WorkerId, wrkItem);

                //Add onto the state, to account for the work just sent.
                //This gives increased accuracy since it could be a while
                //to get an updated heartbeat from this worker with the
                //updated count.
                lock (_workers)
                    _workers[nextWorker.WorkerId].PendingWorkItemCount++;
            }
        }

        private WorkerStatus GetNextWorker()
        {
            //Quickly get the WorkerStatus values in our dictionary.
            List<WorkerStatus> workers = new List<WorkerStatus>();
            lock (_workers)
                workers.AddRange(_workers.Values);

            //Eliminate offline and unknown Workers
            //and sort by the PendingWorkItemCount.
            //Ideally something like a min heap would be a
            //better data structure.  But this is more than
            //sufficient for now.
            return workers
                .Where<WorkerStatus>(
                    w => w.State != WorkerState.Offline
                    && w.State != WorkerState.Unknown)
                .OrderBy<WorkerStatus, int>(w => w.PendingWorkItemCount)
                .FirstOrDefault<WorkerStatus>();
        }
    }
}
