﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.ServiceModel;
using System.Threading;
using PeerWcfComm;

namespace WorkerPeerApplication
{
    /// <summary>
    /// Defintion of Worker peer. It participates in IWorker and IWorkBackup communication.
    /// </summary>
    [ServiceBehavior(InstanceContextMode = InstanceContextMode.Single)]
    public class Worker : IWorker, IWorkBackup
    {
        #region Private Fields...

        private ServiceHost _host;

        private ChannelFactory<IWorker> _workerFactory;
        private ChannelFactory<ISpectator> _spectatorFactory;
        private ChannelFactory<IWorkBackup> _backupFactory;
        private IWorker _workerClientChannel;
        private ISpectator _spectatorClientChannel;
        private IWorkBackup _backupClientChannel;

        private Guid _id;
        private volatile WorkerState _state;
        private volatile WorkItem _currentWorkItem;

        private Queue<WorkItem> _workQueue;
        private Dictionary<Guid, WorkerStatus> _workers;
        private Dictionary<Guid, WorkItemBackupRequest> _workItemBackups;

        private Timer _heartbeat;
        private Timer _backupMonitor;

        #endregion

        #region Public Ctors and Methods...
        public Worker()
        {
            _id = Guid.NewGuid();
            _state = WorkerState.Offline;
            _workQueue = new Queue<WorkItem>();
            _workers = new Dictionary<Guid, WorkerStatus>();
            _workItemBackups = new Dictionary<Guid, WorkItemBackupRequest>();
        }

        public void Start()
        {
            _host = new ServiceHost(this);
            _host.Open();

            _workerFactory = new ChannelFactory<IWorker>("IWorkerClient");
            _workerClientChannel = _workerFactory.CreateChannel();

            _spectatorFactory = new ChannelFactory<ISpectator>("ISpectatorClient");
            _spectatorClientChannel = _spectatorFactory.CreateChannel();

            _backupFactory = new ChannelFactory<IWorkBackup>("IWorkBackupClient");
            _backupClientChannel = _backupFactory.CreateChannel();

            PeerNode pn = ((IClientChannel)_workerClientChannel).GetProperty<PeerNode>();
            pn.MessagePropagationFilter = new RemoteOnlyMessagePropagationFilter();

            ThreadPool.QueueUserWorkItem(MonitorWork);

            _state = WorkerState.Ready;
            _heartbeat = new Timer(
                this.SendHeartbeat,
                null,
                new TimeSpan(0),
                new TimeSpan(0, 0, 1)); //Time should be read in from config.

            _backupMonitor = new Timer(
                this.MonitorBackups,
                null,
                new TimeSpan(0),
                new TimeSpan(0, 1, 0)); //Time should be read in from config.
        }

        public void Stop()
        {
            _state = WorkerState.Offline;
            _heartbeat.Dispose();
            _backupMonitor.Dispose();
            this.SendHeartbeat(null);
            _workerFactory.Close();
            _spectatorFactory.Close();
            _backupFactory.Close();
            _host.Close();
        }
        #endregion

        #region Private Methods...
        private void SendHeartbeat(object context)
        {
            WorkerStatus ws = new WorkerStatus();
            ws.WorkerId = _id;
            ws.State = _state;
            ws.PendingWorkItemCount = _workQueue.Count;
            ws.CurrentWorkItem =
                (_currentWorkItem == null)
                ? Guid.Empty
                : _currentWorkItem.WorkItemId;

            _workerClientChannel.Heartbeat(ws);
        }

        private void MonitorWork(object context)
        {
            if (_workQueue.Count > 0)
            {
                lock (_workQueue)
                    _currentWorkItem = _workQueue.Dequeue();

                DoWorkInternal(_currentWorkItem);
                _currentWorkItem = null;
            }

            if (_state != WorkerState.Offline)
            {
                if (_workQueue.Count > 1)
                {
                    WorkerStatus firstIdleWorker = GetNextIdleWorker();

                    if (firstIdleWorker != null)
                    {
                        WorkItem workToForward;

                        lock (_workQueue)
                            workToForward = _workQueue.Dequeue();

                        _workerClientChannel.DoWork(
                            firstIdleWorker.WorkerId,
                            workToForward);
                    }
                }

                Thread.Sleep(100); //Brief pause to not spin up CPU cycles!
                ThreadPool.QueueUserWorkItem(MonitorWork);
            }
        }

        private void MonitorBackups(object context)
        {
            List<WorkItemBackupRequest> backups = new List<WorkItemBackupRequest>();
            TimeSpan timeOutThreshold = new TimeSpan(0, 1, 0); //This should be read in from config.

            lock (_workItemBackups)
            {
                backups.AddRange(_workItemBackups.Values);

                foreach (WorkItemBackupRequest backup in backups)
                {
                    //If we've passed the RestoreAt time, we need to
                    //take on this work item.
                    if (backup.RestoreAt < DateTime.Now)
                    {
                        lock (_workQueue)
                            _workQueue.Enqueue(backup.Item);

                        _workItemBackups.Remove(backup.Item.WorkItemId);

                        Console.WriteLine("Restored item '{0}'", backup.Item.WorkItemId);
                    }

                    //This work item has not been deleted yet. This is suspicious
                    //so investigate further.
                    else if (backup.BackupRequestTimeStamp.Add(timeOutThreshold) > DateTime.Now)
                    {
                        //Check to see if the original worker is still operating.
                        //If so, assume that either:
                        //  1. It's busy and just hasn't gotten to this yet.
                        //  2. It did process this but somehow didn't tell us
                        //      to delete the backup.
                        //Regardless, we won't do anything and the backup
                        //remains in our cache.
                        if (!WorkerIsStillActive(backup.OriginalWorkerId))
                        {
                            //The machine doesn't seem to be actively working.
                            //Set a deadline for restoration of the work item.
                            backup.RestoreAt = DateTime.Now.Add(timeOutThreshold);

                            //Tell the world our intentions to restore this.
                            _backupClientChannel.RestoringBackup(this._id, backup.Item.WorkItemId);

                            Console.WriteLine("Planning to restore '{0}'", backup.Item.WorkItemId);
                        }
                    }
                }
            }
        }

        private bool WorkerIsStillActive(Guid workerId)
        {
            bool active = false;

            lock (_workers)
            {
                if (_workers.ContainsKey(workerId))
                {
                    WorkerStatus sts = _workers[workerId];
                    if (sts.State == WorkerState.Busy)
                        active = true;
                    else if (sts.State == WorkerState.Ready && sts.PendingWorkItemCount > 0)
                        active = true;
                }
            }

            return active;
        }

        private WorkerStatus GetNextIdleWorker()
        {
            List<WorkerStatus> workers = new List<WorkerStatus>();

            lock (_workers)
                workers.AddRange(_workers.Values);

            return workers
                .Where<WorkerStatus>(
                    w =>
                        (w.State == WorkerState.Ready
                        && w.PendingWorkItemCount == 0))
                .FirstOrDefault<WorkerStatus>();
        }

        private void DoWorkInternal(WorkItem wrk)
        {
            this._state = WorkerState.Busy;
            _spectatorClientChannel.WorkItemStateChanged(
                new WorkItemStatus()
                {
                    WorkerId = _id,
                    WorkItemId = wrk.WorkItemId,
                    NewState = WorkItemState.Executing
                });

            //Doing our work...
            Thread.Sleep(new TimeSpan(0, 0, wrk.PretendToWorkSeconds));

            this._backupClientChannel.DeleteBackup(wrk.WorkItemId);

            this._state = WorkerState.Ready;
            _spectatorClientChannel.WorkItemStateChanged(
                new WorkItemStatus()
                {
                    WorkerId = _id,
                    WorkItemId = wrk.WorkItemId,
                    NewState = WorkItemState.Finished
                });

        }
        #endregion

        #region IWorker Members

        void IWorker.Heartbeat(WorkerStatus status)
        {
            lock (_workers)
                _workers[status.WorkerId] = status;
        }

        void IWorker.DoWork(Guid workerId, WorkItem work)
        {
            if (workerId == _id)
            {
                lock (_workQueue)
                    _workQueue.Enqueue(work);

                this._backupClientChannel.BackupItem(
                    new WorkItemBackupRequest()
                    {
                        Item = work,
                        OriginalWorkerId = _id
                    });
            }
        }

        #endregion

        #region IWorkBackup Members

        void IWorkBackup.BackupItem(WorkItemBackupRequest request)
        {
            request.BackupRequestTimeStamp = DateTime.Now;
            request.RestoreAt = DateTime.MaxValue;

            lock (_workItemBackups)
                _workItemBackups.Add(request.Item.WorkItemId, request);
        }

        void IWorkBackup.DeleteBackup(Guid workItemId)
        {
            lock (_workItemBackups)
                if (_workItemBackups.ContainsKey(workItemId))
                    _workItemBackups.Remove(workItemId);
        }

        void IWorkBackup.RestoringBackup(Guid workerId, Guid workItemId)
        {
            lock (_workItemBackups)
            {
                //Only proceed if we have this item backed up.
                if (_workItemBackups.ContainsKey(workItemId))
                {
                    //Get the backup request for this work item
                    WorkItemBackupRequest req = _workItemBackups[workItemId];

                    //Check to see if we've decided to restore this.
                    if (req.RestoreAt != DateTime.MaxValue)
                    {
                        //The restorer with the greatest id wins!
                        if (this._id.CompareTo(workItemId) < 0)
                            _workItemBackups.Remove(workItemId);
                    }
                    else
                    {
                        //We haven't decided to restore it and someone else wants to.
                        //So let them.
                        _workItemBackups.Remove(workItemId);
                    }
                }
            }
        }

        #endregion
    }
}
